/***************************************************************************
 *   Copyright (C) 2005-2006 Nicolas Hadacek <hadacek@kde.org>             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "pic_memory.h"

#include <qfile.h>

#include "common/common/misc.h"

Pic::Memory::Memory(const Data &data)
  : Device::Memory(data), _ranges(Pic::Nb_MemoryRangeTypes)
{
  for (uint i=0; i<Nb_MemoryRangeTypes; i++) {
    uint nb = device().nbWords(MemoryRangeType(i));
    _ranges[i].resize(nb);
  }
  fill(Device::INVALID);
}

void Pic::Memory::fill(MemoryRangeType type, uint value)
{
  for (uint i=0; i<_ranges[type].count(); i++) {
    if ( type==MemoryConfig && value==Device::INVALID ) _ranges[type][i] = device().config()._words[i].bvalue;
    else _ranges[type][i] = value;
  }
}

bool Pic::Memory::isClear(MemoryRangeType type) const
{
  for (uint i=0; i<_ranges[type].count(); i++) {
    if ( type==MemoryConfig ) {
      if ( _ranges[type][i]!=device().config()._words[i].bvalue ) return false;
    } else if (_ranges[type][i]!=Device::INVALID ) return false;
  }
  return true;
}

void Pic::Memory::fill(uint value)
{
  for (uint k=0; k<Nb_MemoryRangeTypes; k++) fill(MemoryRangeType(k), value);
}

void Pic::Memory::copyFrom(MemoryRangeType type, const Memory &memory)
{
  Q_ASSERT( memory.device().name()==device().name() );
  for (uint i=0; i<_ranges[type].count(); i++) _ranges[type][i] = memory._ranges[type][i];
}

void Pic::Memory::copyFrom(const Device::Memory &memory)
{
  Q_ASSERT( memory.device().name()==device().name() );
  for (uint i=0; i<Nb_MemoryRangeTypes; i++)
    copyFrom(MemoryRangeType(i), static_cast<const Memory &>(memory));
}

Device::Array Pic::Memory::arrayForWriting(MemoryRangeType type) const
{
  Device::Array data = _ranges[type];
  for (uint i=0; i<data.count(); i++)
    data[i] &= (type==MemoryConfig ? device().config()._words[i].wmask : device().mask(type));
  return data;
}

uint Pic::Memory::word(MemoryRangeType type, uint offset) const
{
  if ( offset>=_ranges[type].size() ) qCrash();
  return _ranges[type][offset];
}

uint Pic::Memory::normalizeWord(MemoryRangeType type, uint offset, uint value) const
{
  if ( type==MemoryConfig) {
    const Config::Word &cword = device().config()._words[offset];
    return value & cword.usedMask();
  }
  uint mask = device().range(type).rmask; // rmask is used in User Id
  if ( (value & mask)!=value ) return mask;
  return value;
}

uint Pic::Memory::normalizedWord(MemoryRangeType type, uint offset) const
{
  return normalizeWord(type, offset, word(type, offset));
}

void Pic::Memory::setWord(MemoryRangeType type, uint offset, uint value)
{
  Q_ASSERT( _ranges[type].size()>offset );
  //qDebug("Memory::setWord at %s %i: %s", MEMORY_RANGE_TYPE_DATA[type].label, offset, toHex(value, 4).data());
  _ranges[type][offset] = value;
}

void Pic::Memory::setArray(MemoryRangeType type, const Device::Array &data)
{
  Q_ASSERT( _ranges[type].size()==data.size() );
  _ranges[type] = data;
}

Range Pic::Memory::bootRange() const
{
  // with BBSIZ
  uint i;
  const Config::Mask *mask = device().config().findMask("BBSIZ", &i);
  if (mask) {
    uint v = word(Pic::MemoryConfig, i) & mask->value;
    for (uint k=0; k<uint(mask->values.count()); k++) {
      if ( (v & mask->values[k].value)==v )
        return Range(0, 2 * mask->values[k].name.toUInt() - 1); // words
    }
    Q_ASSERT(false);
  }

  // only CPB
  QString maskName = device().config().maskName(Config::ProgramProtected, Config::BootBlock);
  mask = device().config().findMask(maskName);
  for (uint k=0; k<uint(mask->values.count()); k++) {
    Range range = device().config().extractRange(mask->values[k].name, Config::BootBlock);
    if ( !range.isEmpty() ) return range;
  }
  Q_ASSERT(false);
  return Range();
}

Range Pic::Memory::blockRange(uint i) const
{
  Q_ASSERT( i<device().config().nbBlocks() );
  Range range = (i==0 ? bootRange() : blockRange(i-1));
  QString maskName = device().config().maskName(Config::ProgramProtected, Config::CodeBlock, i);
  const Config::Mask *mask = device().config().findMask(maskName);
  for (uint k=0; k<uint(mask->values.count()); k++) {
    Range r = device().config().extractRange(mask->values[k].name, Config::CodeBlock);
    if ( !r.isEmpty() ) return Range(range.end+1, r.end);
  }
  Q_ASSERT(false);
  return Range();
}

bool Pic::Memory::isProtected(Config::ProtectionType ptype, Config::BlockType type, int block) const
{
  return !protectedRange(ptype, type, block).isEmpty();
}

bool Pic::Memory::isProtected(Config::ProtectionType ptype, MemoryRangeType type) const
{
  RangeVector rv = protectedRanges(ptype, type);
  return !rv.isEmpty();
}

RangeVector Pic::Memory::protectedRanges(Config::ProtectionType ptype, MemoryRangeType type) const
{
  RangeVector rv;
  switch (type) {
    case MemoryCode:
      if ( device().config().hasBootBlock() && isProtected(ptype, Config::BootBlock) ) rv.append(bootRange());
      if ( device().config().nbBlocks()!=0 ) {
        for (uint i=0; i<device().config().nbBlocks(); i++)
          if ( isProtected(ptype, Config::CodeBlock, i) ) rv.append(blockRange(i));
      } else rv.append(protectedRange(ptype, Config::CodeBlock, -1));
      break;
    case MemoryEeprom:
      rv.append(protectedRange(ptype, Config::DataBlock, -1));
      break;
    case MemoryConfig:
      rv.append(protectedRange(ptype, Config::ConfigBlock, -1));
      break;
    case MemoryCal:
      rv.append(protectedRange(ptype, Config::CalibrationBlock, -1));
      break;
    case MemoryUserId:
    case MemoryDeviceId:
    case MemoryCalBackup:
    case MemoryDebugVector:
    case MemoryProgramExecutive:
    case MemoryHardwareStack: break;
    case Nb_MemoryRangeTypes: Q_ASSERT(false); break;
  }
  return rv;
}

Range Pic::Memory::protectedRange(Config::ProtectionType ptype, Config::BlockType type, int block) const
{
  QString maskName = device().config().maskName(ptype, type, block);
  if ( maskName.isEmpty() ) return Range();
  uint i;
  const Config::Mask *mask = device().config().findMask(maskName, &i);
  if ( mask==0 ) return Range();
  uint v = word(Pic::MemoryConfig, i) & mask->value;
  for (uint k=0; k<uint(mask->values.count()); k++) {
    if ( (v & mask->values[k].value)==v )
      return device().config().extractRange(mask->values[k].name, type);
  }
  Q_ASSERT(false);
  return Range();
}

void Pic::Memory::setProtection(bool on, Config::ProtectionType ptype, Config::BlockType type, int block)
{
  QString maskName = device().config().maskName(ptype, type, block);
  uint i;
  const Config::Mask *mask = device().config().findMask(maskName, &i);
  if( mask==0 ) return;
  uint v = word(Pic::MemoryConfig, i);
  v &= ~mask->value; // clear mask bits
  for (uint k=0; k<uint(mask->values.count()); k++) {
    if ( (!on && mask->values[k].name==device().config().noneProtectedValueName())
         || (on && mask->values[k].name=="All") )
      setWord(MemoryConfig, i, v | mask->values[k].value);
  }
}

bool Pic::Memory::hasFlagOn(const QString &name) const
{
  uint i;
  const Config::Mask *mask = device().config().findMask(name, &i);
  if ( mask==0 ) return false;
  uint v = word(Pic::MemoryConfig, i) & mask->value;
  for (uint k=0; k<uint(mask->values.count()); k++)
    if ( (v & mask->values[k].value)==v ) return mask->values[k].name=="On";
  Q_ASSERT(false);
  return false;
}

void Pic::Memory::setFlagOn(const QString &name, bool on)
{
  uint i;
  const Config::Mask *mask = device().config().findMask(name, &i);
  Q_ASSERT(mask);
  uint v = word(Pic::MemoryConfig, i);
  v &= ~mask->value; // clear mask bits
  for (uint k=0; k<uint(mask->values.count()); k++) {
    if ( XOR(on, mask->values[k].name=="On") ) continue;
    setWord(Pic::MemoryConfig, i, v | mask->values[k].value);
    return;
  }
  Q_ASSERT(false);
}

uint Pic::Memory::checksum() const
{
  uint cs = 0;
  // code
  uint mask = device().mask(MemoryCode);
  RangeVector rv = protectedRanges(Config::ProgramProtected, MemoryCode);
  //qDebug("rv empty: %i  nb: %i (%s)", rv.isEmpty(), device().nbWords(MemoryCode), toHex(mask, 4).data());
  //for (uint i=0; i<rv.count(); i++)
  //  qDebug("protected: %s-%s", toHex(rv[i].start, 4).data(), toHex(rv[i].end, 4).data());
  for (uint i=0; i<device().nbWords(MemoryCode); i++) {
    if ( rv.contains(i) ) continue;
    uint v = word(MemoryCode, i) & mask;
    switch ( device().architecture() ) {
      case Arch_18X: // #### not true for all 18F ??
        cs += (v & 0xFF) + (v >> 8);
        break;
      case Arch_30X:
        cs += (v & 0xFF) + ((v >> 8) & 0xFF) + (v >> 16);
        break;
      default:
        cs += v;
        break;
    }
  }
  //qDebug("code: %s", toHex(cs, 9).data());
  // config
  const Config &config = device().config();
  for (uint i=0; i<uint(config._words.count()); i++) {
    const Config::Word &cword = config._words[i];
    uint v = word(MemoryConfig, i) & cword.cmask;
    //qDebug("%i: %s %s", i, toHex(word(MemoryConfig, i), 4).data(), toHex(cword.cmask, 4).data());
    if ( device().architecture()==Arch_30X ) cs += (v & 0xFF) + (v >>8);
    else cs += v;
  }
  const ChecksumData &cdata = device().checksumData();
  cs += (rv.isEmpty() ? cdata.constant : cdata.protectedConstant);
  //qDebug("%s", toHex(cs, 9).data());
  // user ids
  if ( !rv.isEmpty() && device().isPresent(MemoryUserId) ) {
    uint id = 0;
    uint nb = device().nbWords(MemoryUserId);
    uint nbb = (device().name()=="14000" ? 7 : 4);
    for (uint i=0; i<nb; i++) {
      uint v = word(MemoryUserId, nb-i-1) & 0xF;
      if ( device().architecture()==Arch_18X ) id += v;
      else {
        //qDebug("id %i (%i): %s %s", i, nbb, toHex(v, 4).data(), toHex(v << (nbb*i), 9).data());
        id += v << (nbb*i);
      }
    }
    //qDebug("id %s", toHex(id, 9).data());
    cs += id;
  }
  //qDebug("%s %s", toHex(cs, 9).data(), toHex(cs & 0xFFFF, 4).data());
  return cs & 0xFFFF;
}

uint Pic::Memory::unprotectedChecksum() const
{
  Memory tmp(*this);
  if ( device().config().hasBootBlock() )
    tmp.setProtection(false, Config::ProgramProtected, Config::BootBlock, -1);
  if ( device().config().nbBlocks()!=0 ) {
    for (uint i=0; i<device().config().nbBlocks(); i++)
      tmp.setProtection(false, Config::ProgramProtected, Config::CodeBlock, i);
  } else tmp.setProtection(false, Config::ProgramProtected, Config::CodeBlock, -1);
  if ( device().isReadable(MemoryEeprom) ) tmp.setProtection(false, Config::ProgramProtected, Config::DataBlock, -1);
  tmp.setProtection(false, Config::ProgramProtected, Config::ConfigBlock, -1);
  tmp.setProtection(false, Config::ProgramProtected, Config::CalibrationBlock, -1);
  return tmp.checksum();
}

void Pic::Memory::setUserIdToUnprotectedChecksum()
{
  uint cs = unprotectedChecksum();
  uint nb = device().nbWords(MemoryUserId);
  for (uint i=0; i<nb; i++) setWord(MemoryUserId, nb-i-1, (cs >> (4*i)) & 0xF);
}

//-----------------------------------------------------------------------------
void Pic::Memory::savePartial(QTextStream &stream, HexBuffer::Format format) const
{
  // save memory ranges in the same order as MPLAB (for easy checking)
  const MemoryRangeType saveOrder[] = { MemoryCode, MemoryEeprom, MemoryConfig, MemoryUserId, MemoryCal, Nb_MemoryRangeTypes };
  HexBuffer hb;
  for (uint i=0; saveOrder[i]!=Nb_MemoryRangeTypes; i++) {
    hb.clear();
    toHexBuffer(saveOrder[i], hb);
    hb.savePartial(stream, format);
  }
}

//-----------------------------------------------------------------------------
void Pic::Memory::toHexBuffer(MemoryRangeType type, HexBuffer &hb) const
{
  if ( !device().isWritable(type) ) return;
  uint nbBytes = ARCHITECTURE_DATA[device().architecture()].nbBytesWord;
  bool packed = ARCHITECTURE_DATA[device().architecture()].packed;
  uint offset = device().range(type).hexFileOffset;
  if ( offset==0 ) offset = device().range(type).start;
  uint mask = device().mask(type);
  uint wNbBytes = nbBytes;
  if ( packed && type!=Pic::MemoryCode ) {
    offset /= 2;
    wNbBytes /= 2;
  }
  uint byte = 0;
  uint wOffset = 0;
  uint wByte = 0;
  //qDebug("%s wnb=%i snb=%i div=%i", MEMORY_RANGE_TYPE_DATA[type].label, wNbBytes, sNbBytes, div);
  for (uint k=0; k<wNbBytes*device().nbWords(type); k++) {
    // set byte
    uint s = _ranges[type][wOffset] & mask;
    //if ( k<4 ) qDebug("s=%s so=%s sb=%i wo=%i wb=%i", toHex(s, 8).data(), toHex(sOffset, 8).data(), sByte, wOffset, wByte);
    s = s >> (wByte*8);
    s = s & 0xFF;
    if ( (byte%2)==0 ) hb.set(offset, s);
    else hb.set(offset, hb[offset] | (s << ((byte%2)*8)));
    // increment offsets
    byte++;
    if ( (byte%nbBytes)==0 ) {
      byte = 0;
      offset++;
    } else if ( byte==2 ) offset++;
    wByte++;
    if ( (wByte%wNbBytes)==0 ) {
      wByte = 0;
      wOffset++;
    }
  }
}

HexBuffer Pic::Memory::toHexBuffer() const
{
  HexBuffer hb;
  for (uint i=0; i<Nb_MemoryRangeTypes; i++) toHexBuffer(MemoryRangeType(i), hb);
  return hb;
}

void Pic::Memory::fromHexBuffer(MemoryRangeType type, const HexBuffer &hb, WarningTypes &result,
                                QStringList &warnings, QMap<uint, bool> &inRange)
{
  if ( !device().isWritable(type) ) return;
  uint nbBytes = ARCHITECTURE_DATA[device().architecture()].nbBytesWord;
  bool packed = ARCHITECTURE_DATA[device().architecture()].packed;
  uint offset = device().range(type).hexFileOffset;
  if ( offset==0 ) offset = device().range(type).start;
  uint mask = device().mask(type);
  uint wNbBytes = nbBytes;
  if ( packed && type!=Pic::MemoryCode ) {
    offset /= 2;
    wNbBytes /= 2;
  }
  uint byte = 0;
  uint wOffset = 0;
  uint wByte = 0;
  //qDebug("%s wnb=%i snb=%i", MEMORY_RANGE_TYPE_DATA[type].label, wNbBytes, nbBytes);
  for (uint k=0; k<wNbBytes*device().nbWords(type); k++) {
    // set byte
    uint s = hb[offset];
    //if ( k<4 ) qDebug("s=%s so=%s sb=%i wo=%i wb=%i", toHex(s, 8).data(), toHex(offset, 8).data(), byte, wOffset, wByte);
    if ( s==Device::INVALID ) {
      if ( type==MemoryConfig ) _ranges[type][wOffset] = mask;
      else _ranges[type][wOffset] = Device::INVALID;
    } else {
      inRange[offset] = true;
      s = s >> ((byte%2)*8);
      s = s & 0xFF;
      if ( wByte==0 ) _ranges[type][wOffset] = s;
      else _ranges[type][wOffset] |= (s << (wByte*8));
    }
    // increment offsets
    byte++;
    if ( (byte%nbBytes)==0 ) {
      byte = 0;
      offset++;
    } else if ( byte==2 ) offset++;
    wByte++;
    if ( (wByte%wNbBytes)==0 ) {
      if ( _ranges[type][wOffset]!=Device::INVALID ) {
        if ( !(result & ValueTooLarge) && (_ranges[type][wOffset] & mask)!=_ranges[type][wOffset] ) {
          result |= ValueTooLarge;
          warnings += i18n("At least one word (at offset %1) is larger (%2) than the corresponding mask (%3).")
                      .arg(toHexLabel(offset, 8)).arg(toHexLabel(_ranges[type][wOffset], 8)).arg(toHexLabel(mask, 8));
        }
        _ranges[type][wOffset] = (_ranges[type][wOffset] & mask);
      }
      wByte = 0;
      wOffset++;
    }
  }
}

void Pic::Memory::fromHexBuffer(const HexBuffer &hb, WarningTypes &result,
                                QStringList &warnings, QMap<uint, bool> &inRange)
{
  for (uint i=0; i<Nb_MemoryRangeTypes; i++)
    fromHexBuffer(MemoryRangeType(i), hb, result, warnings, inRange);
}
