/***************************************************************************
 *   Copyright (C) 2007 Nicolas Hadacek <hadacek@kde.org>                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "bootloader_config_widget.h"

#include "progs/bootloader/base/bootloader_prog.h"
#include "progs/bootloader/base/bootloader_tbl.h"
#include "common/port/serial.h"

//-----------------------------------------------------------------------------
TinyBootloader::ConfigWidget::ConfigWidget()
  : GridConfigWidget(0, "tiny_bootloader_config_widget")
{
  uint row = _topLayout->numRows();

  QLabel *label = new QLabel(i18n("Port Speed:"), this);
  _topLayout->addWidget(label, row, 0);
  _speed = new KComboBox(this);
  for (uint i=0; i<Port::Serial::Nb_Speeds; i++) {
    if ( Port::Serial::SPEED_VALUES[i]==0 || !Port::Serial::SPEED_DATA[i].supported ) continue;
    _speed->insertItem(QString::number(Port::Serial::SPEED_VALUES[i]));
  }
  _topLayout->addWidget(_speed, row, 1);
  row++;

  label = new QLabel(i18n("Timeout (ms):"), this);
  _topLayout->addWidget(label, row, 0);
  _timeout = new KIntNumInput(this);
  _timeout->setMinValue(0);
  _topLayout->addWidget(_timeout, row, 1);
  row++;

  label = new QLabel(i18n("No of Retries:"), this);
  _topLayout->addWidget(label, row, 0);
  _retries = new KIntNumInput(this);
  _retries->setMinValue(0);
  _topLayout->addWidget(_retries, row, 1);
  row++;

  _topLayout->setColStretch(2, 1);
}

void TinyBootloader::ConfigWidget::saveConfig()
{
  Config config;
  uint k = 0;
  for (uint i=0; i<Port::Serial::Nb_Speeds; i++) {
    if ( Port::Serial::SPEED_VALUES[i]==0 || !Port::Serial::SPEED_DATA[i].supported ) continue;
    if ( uint(_speed->currentItem())==k ) {
      config.writeSpeed(Port::Serial::Speed(i));
      break;
    }
    k++;
  }
  config.writeTimeout(_timeout->value());
  config.writeRetries(_retries->value());
}

void TinyBootloader::ConfigWidget::loadConfig()
{
  Config config;
  Port::Serial::Speed speed = config.readSpeed();
  uint k = 0;
  for (uint i=0; i<Port::Serial::Nb_Speeds; i++) {
    if ( Port::Serial::SPEED_VALUES[i]==0 || !Port::Serial::SPEED_DATA[i].supported ) continue;
    if ( i==uint(speed) ) break;
    k++;
  }
  _speed->setCurrentItem(k);
  _timeout->setValue(config.readTimeout());
  _retries->setValue(config.readRetries());
}

//-----------------------------------------------------------------------------
::Programmer::ConfigWidget *Bootloader::GroupUI::createConfigWidget(QWidget *parent) const
{
  return new ConfigWidget(static_cast<const ::Programmer::Group &>(group()), parent);
}

//-----------------------------------------------------------------------------
Bootloader::ConfigWidget::ConfigWidget(const ::Programmer::Group &group, QWidget *parent)
  : ::Programmer::ConfigWidget(group, parent, "bootloader_config_widget")
{
  uint row = _topLayout->numRows();

  QLabel *label = new QLabel(i18n("Bootloader:"), this);
  _topLayout->addWidget(label, row, 0);
  _type = new KComboBox(this);
  for (uint i=0; i<Nb_Types; i++) _type->insertItem(i18n(TYPE_DATA[i].label)); // to get the maximum size
  connect(_type, SIGNAL(activated(int)), SLOT(typeChanged()));
  _topLayout->addWidget(_type, row, 1);
  row++;

  _eeprom = new QCheckBox(i18n("Program EEPROM (if supported)"), this);
  _topLayout->addMultiCellWidget(_eeprom, row,row, 0,1);
  row++;

  _stack = new QWidgetStack(this);
  for (uint i=0; i<Nb_Types; i++) {
    ::ConfigWidget *cw = createConfigWidget(Type(i));
    _stack->addWidget(cw, i);
    cw->loadConfig();
  }
  _topLayout->addMultiCellWidget(_stack, row,row, 0, 1);
  row++;

  _topLayout->setColStretch(2, 1);
}

ConfigWidget *Bootloader::ConfigWidget::createConfigWidget(Type type)
{
  switch (type) {
    case TinyBootloader: return new ::TinyBootloader::ConfigWidget;
    case Pickit2Bootloader: return new ::Pickit2Bootloader::ConfigWidget;
    case PicdemBootloader: return new ::PicdemBootloader::ConfigWidget;
    case Nb_Types: break;
  }
  Q_ASSERT(false);
  return 0;
}

bool Bootloader::ConfigWidget::setPort(const Port::Description &pd)
{
  _portType = pd.type;
  _type->clear();
  for (uint i=0; i<Nb_Types; i++) {
    if ( TYPE_DATA[i].portType!=pd.type ) continue;
    _type->insertItem(i18n(TYPE_DATA[i].label));
  }
  _type->setCurrentItem(0);
  typeChanged();
  return ::Programmer::ConfigWidget::setPort(pd);
}

Bootloader::Type Bootloader::ConfigWidget::type() const
{
  int k = 0;
  for (uint i=0; i<Nb_Types; i++) {
    if ( TYPE_DATA[i].portType!=_portType ) continue;
    if ( k==_type->currentItem() ) return Type(i);
    k++;
  }
  Q_ASSERT(false);
  return Nb_Types;
}

void Bootloader::ConfigWidget::typeChanged()
{
  _stack->raiseWidget(type());
}

void Bootloader::ConfigWidget::saveConfig()
{
  ::Programmer::ConfigWidget::saveConfig();
  Config config(static_cast<const Group &>(_group));
  config.writeType(Type(_type->currentItem()));
  config.writeProgramEeprom(_eeprom->isChecked());
  static_cast< ::ConfigWidget *>(_stack->visibleWidget())->saveConfig();
}

void Bootloader::ConfigWidget::loadConfig()
{
  ::Programmer::ConfigWidget::loadConfig();
  Config config(static_cast<const Group &>(_group));
  _type->setCurrentItem(config.readType());
  _eeprom->setChecked(config.readProgramEeprom());
  if ( _stack->visibleWidget() ) static_cast< ::ConfigWidget *>(_stack->visibleWidget())->loadConfig();
}
